import os
import signal
import subprocess
import psutil

import txtapi.__main__
from txtapi.environment.environment import PID_FILENAME, WORKSPACE_BASEPATH
from txtapi.handlers import workspace_handler
from txtapi.handlers import util_handler
# pylint: disable=no-name-in-module, import-error
from txtapi.utility.watch_process import WatchProcess
#pylint: enable=no-name-in-module, import-error

def start_application(workspace_name):
    
    try:
        util_handler.stop()
    except:
        pass
    
    executablePath = workspace_handler.get_file_path(workspace_name, workspace_name + ".py")
    if not os.path.exists(executablePath):
        return "Failed to start process: " + executablePath + " does not exist."

    try:
        # -u: force the stdout and stderr streams to be unbuffered;
        process = subprocess.Popen(["/usr/bin/python3", "-u", executablePath], env=os.environ.copy(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        watch_process = WatchProcess.get_instance()
        watch_process.start(process)

        print("Starting process with pid " + str(process.pid))
        pid_filepath = workspace_handler.get_pid_file_path(workspace_name)
        print("saving PID in " + pid_filepath)
        pidFile = open(pid_filepath, "w")  # approot/.pid contains the pid
        pidFile.write(str(process.pid))
        pidFile.close()
        
        if not os.path.isfile(pid_filepath):
            print("pid has not been saved")
        else:
            print("pid file saved")
            
    except Exception as ex:
        return "Failed to start process: " + str(ex)
    
    return "Started process with Pid: " + str(process.pid)

def stop_application(workspace_name):
    
    watch_process = WatchProcess.get_instance()
    watch_process.stop()

    try:
        pid = workspace_handler.get_pid(workspace_name)
    except Exception as e:
        return "Exception: pid not found, " + str(e)

    workspace_handler.delete_file_of_workspace(workspace_name, PID_FILENAME)
    process = psutil.Process(int(pid))
    process.terminate()

    # handle process termination to avoid zombies
    # a process may take some time to terminate
    _, alive = psutil.wait_procs([process], timeout=3)
    if alive:
        # try harder if there are still some processes
        for p in alive:
            p.kill()
        _, alive = psutil.wait_procs(alive, timeout=1)

    if alive:
        return "Failed to stop python process. workspace_name: " + workspace_name + "; pid: " + str(pid)

    return "Python process stopped. workspace_name: " + workspace_name + "; pid: " + str(pid)

def stop_applications():
    
    watch_process = WatchProcess.get_instance()
    watch_process.stop()

    # kill all python processes that are running in /opt/ft/workspaces
    current_process = psutil.Process()
    children = current_process.children(recursive=True)
    terminated = []
    if txtapi.__main__.APPLICATION_PID >= 0:
        try:
            p = psutil.Process(txtapi.__main__.APPLICATION_PID)
            terminated.append(p)
            p.terminate()
        except:
            # process cannot be killed or no such process
            pass
    else:
        for p in children:
            try:
                if 'python3' in p.exe() or 'python3' in p.name() or [ True for arg in p.cmdline() if 'python3' in arg ]:
                    if [ True for arg in p.cmdline() if WORKSPACE_BASEPATH in arg ]:
                        terminated.append(p)
                        p.terminate()
            except:
                # process cannot be killed or no such process
                pass

    # handle process termination to avoid zombies
    # a process may take some time to terminate
    _, alive = psutil.wait_procs(terminated, timeout=3)
    if alive:
        # try harder if there are still some processes
        for p in alive:
            p.kill()
        _, alive = psutil.wait_procs(alive, timeout=1)

    if alive:
        return "Failed to stop python processes"

    return "All python processes stopped."

